@@ -1,5 +1,5 @@
 /*
- * Copyright 2018 Advanced Micro Devices, Inc.
+ * Copyright 2024 Advanced Micro Devices, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -21,30 +21,10 @@
  *
  */
 
-#ifndef __DF_V3_6_H__
-#define __DF_V3_6_H__
+#ifndef __DF_V4_15_H__
+#define __DF_V4_15_H__
 
-#include "soc15_common.h"
+extern const struct amdgpu_df_funcs df_v4_15_funcs;
 
-enum DF_V3_6_MGCG {
-	DF_V3_6_MGCG_DISABLE = 0,
-	DF_V3_6_MGCG_ENABLE_00_CYCLE_DELAY = 1,
-	DF_V3_6_MGCG_ENABLE_01_CYCLE_DELAY = 2,
-	DF_V3_6_MGCG_ENABLE_15_CYCLE_DELAY = 13,
-	DF_V3_6_MGCG_ENABLE_31_CYCLE_DELAY = 14,
-	DF_V3_6_MGCG_ENABLE_63_CYCLE_DELAY = 15
-};
+#endif /* __DF_V4_15_H__ */
 
-/* Defined in global_features.h as FTI_PERFMON_VISIBLE */
-#define DF_V3_6_MAX_COUNTERS		4
-
-/* get flags from df perfmon config */
-#define DF_V3_6_GET_EVENT(x)		(x & 0xFFUL)
-#define DF_V3_6_GET_INSTANCE(x)		((x >> 8) & 0xFFUL)
-#define DF_V3_6_GET_UNITMASK(x)		((x >> 16) & 0xFFUL)
-#define DF_V3_6_PERFMON_OVERFLOW	0xFFFFFFFFFFFFULL
-
-extern const struct attribute_group *df_v3_6_attr_groups[];
-extern const struct amdgpu_df_funcs df_v3_6_funcs;
-
-#endif
